
enum tCastlingTargetSquare
  g1
  c1
  g8
  c8
end enum

type tVector
  as integer x
  as integer y
end type

dim shared as tVector pMove(1 to 6)
dim shared as tVector nMove(1 to 8)
dim shared as tVector bMove(1 to 4)
dim shared as tVector rMove(1 to 4)
dim shared as tVector qMove(1 to 8)
dim shared as tVector kMove(1 to 8)

nMove(1) = type(-1, +2)
nMove(2) = type(+1, +2)
nMove(3) = type(-2, +1)
nMove(4) = type(+2, +1)
nMove(5) = type(-2, -1)
nMove(6) = type(+2, -1)
nMove(7) = type(-1, -2)
nMove(8) = type(+1, -2)
bMove(1) = type(-1, +1)
bMove(2) = type(+1, +1)
bMove(3) = type(-1, -1)
bMove(4) = type(+1, -1)
rMove(1) = type(0, +1)
rMove(2) = type(0, -1)
rMove(3) = type(-1,  0)
rMove(4) = type(+1,  0)

kMove(1) = bMove(1)
kMove(2) = bMove(2)
kMove(3) = bMove(3)
kMove(4) = bMove(4)
kMove(5) = rMove(1)
kMove(6) = rMove(2)
kMove(7) = rMove(3)
kMove(8) = rMove(4)
qMove(1) = kMove(1)
qMove(2) = kMove(2)
qMove(3) = kMove(3)
qMove(4) = kMove(4)
qMove(5) = kMove(5)
qMove(6) = kMove(6)
qMove(7) = kMove(7)
qMove(8) = kMove(8)
pMove(1) = bMove(1)
pMove(2) = bMove(2)
pMove(4) = bMove(3)
pMove(5) = bMove(4)
pMove(3) = rMove(1)
pMove(6) = rMove(2)

StandardPlacement:
data -4,-2,-3,-5,-6,-3,-2,-4
data -1,-1,-1,-1,-1,-1,-1,-1
data  0, 0, 0, 0, 0, 0, 0, 0
data  0, 0, 0, 0, 0, 0, 0, 0
data  0, 0, 0, 0, 0, 0, 0, 0
data  0, 0, 0, 0, 0, 0, 0, 0
data +1,+1,+1,+1,+1,+1,+1,+1
data +4,+2,+3,+5,+6,+3,+2,+4

enum tGameState
  gsInProgress
  gsCheckmate
  gsStalemate
  gsDraw
end enum

namespace chess

function VectToStr(v as const tVector) as string
  return chr(v.x + 96) & chr(v.y + 48)
end function

operator + (v1 as tVector, v2 as tVector) as tVector
  return type(v1.x + v2.x, v1.y + v2.y)
end operator

operator = (v1 as tVector, v2 as tVector) as boolean
  return (v1.x = v2.x) andalso (v1.y = v2.y)
end operator

type tMove
  as tVector a
  as tVector b
  as boolean isLegal
end type

const as integer cNothing = 0
const as integer cPawn    = 1
const as integer cKnight  = 2
const as integer cBishop  = 3
const as integer cRook    = 4
const as integer cQueen   = 5
const as integer cKing    = 6
const as integer cWhite   = 1
const as integer cBlack   =-1

type tChessGame
  as integer board(1 to 8, 1 to 8)
  as integer activeColor
  as boolean castlingAvailable(g1 to c8)
  as tVector enPassantTargetSquare
  as integer halfmoveClock
  as integer fullmoveNumber
  as tMove moves(1 to 100)
  as integer movesNumber
  as tMove opponentMoves(1 to 100)
  as integer opponentMovesNumber
  declare sub Initialize()
  declare function Text() as string
  declare function IsInBoard(as tVector) as boolean
  declare sub GenerateGeometricalMoves()
  declare sub GenerateOpponentGeometricalMoves()
  declare function IsThreatened(as tVector, as boolean = false) as boolean
  declare function Check() as boolean
  declare sub GenerateSpecialMoves()
  declare sub DoMove(as tMove, as integer = cQueen)
  declare sub DoMove(as string, as integer = cQueen)
  declare sub CheckMovesLegality()
  declare function IsLegal(as string) as boolean
  declare function IsLegalStartingSquare(as string) as boolean
  declare function LegalDestinations(as string) as string
  declare function GameState() as tGameState
  declare function RandomMove() as integer
  declare function FEN() as string
  declare function EPD() as string
  declare sub Initialize(as string)
  private:
    declare sub GGM(m() as tMove, byref mn as integer)
end type

sub tChessGame.Initialize()
  dim piece as integer
  restore StandardPlacement
  for y as integer = 8 to 1 step -1
    for x as integer = 1 to 8
      read piece
      board(x, y) = piece
    next x
  next y
  activeColor = cWhite
  for c as tCastlingTargetSquare = g1 to c8
    castlingAvailable(c) = true
  next c
  with enPassantTargetSquare
    .x = 0
    .y = 0
  end with
  halfmoveClock = 0
  fullmoveNumber = 1
end sub

sub tChessGame.Initialize(aFEN as string)
  dim as string s(1 to 6), symbol = "kqrbnp PNBRQK", s1 = aFEN
  dim as integer x, y, z
  
  z = instr(s1, " ") : s(1) = left(s1, z - 1) : s1 = right(s1, len(s1) - z)
  z = instr(s1, " ") : s(2) = left(s1, z - 1) : s1 = right(s1, len(s1) - z)
  z = instr(s1, " ") : s(3) = left(s1, z - 1) : s1 = right(s1, len(s1) - z)
  z = instr(s1, " ") : s(4) = left(s1, z - 1) : s1 = right(s1, len(s1) - z)
  z = instr(s1, " ") : s(5) = left(s1, z - 1) : s1 = right(s1, len(s1) - z)
  s(6) = s1
  
  x = 1
  y = 8
  for z = 1 to len(s(1))
    if s(1)[z - 1] = asc("/") then
      x = 1
      y -= 1
      continue for
    end if
    if (s(1)[z - 1] >= asc("1")) andalso (s(1)[z - 1] <= asc("8")) then
      do while s(1)[z - 1] > asc("0")
        board(x, y) = cNothing
        s(1)[z - 1] -= 1
        x += 1
      loop
      continue for
    end if
    if instr(symbol, chr(s(1)[z - 1])) > 0 then
      board(x, y) = instr(symbol, chr(s(1)[z - 1])) - 7
      x += 1
      continue for
    end if
  next z
  
  activeColor = iif(s(2) = "w", cWhite, cBlack)
  
  if instr(s(3), "K") > 0 then castlingAvailable(g1) = true else castlingAvailable(g1) = false
  if instr(s(3), "Q") > 0 then castlingAvailable(c1) = true else castlingAvailable(c1) = false
  if instr(s(3), "k") > 0 then castlingAvailable(g8) = true else castlingAvailable(g8) = false
  if instr(s(3), "q") > 0 then castlingAvailable(c8) = true else castlingAvailable(c8) = false
  
  if len(s(4)) = 2 then
    enPassantTargetSquare.x = s(4)[0] - 96
    enPassantTargetSquare.y = s(4)[1] - 48
  else
    enpassantTargetSquare.x = 0
    enPassantTargetSquare.y = 0
  end if
  
  halfmoveClock = valint(s(5))
  fullmoveNumber = valint(s(6))
end sub

function tChessGame.Text() as string
  const letters = "+ A B C D E F G H"
  dim symbol as string = "kqrbnp-PNBRQK"
  dim result as string = letters & chr(13, 10)
  for y as integer = 8 to 1 step -1
    result += chr(y + 48)
    for x as integer = 1 to 8
      result += chr(32, symbol[board(x, y) + cKing])
    next x
    result += chr(13, 10)
  next y
  result += letters
  function = result
end function

function tChessgame.IsInBoard(v as tVector) as boolean
  return (v.x >= 1) andalso (v.x <= 8) andalso (v.y >= 1) andalso (v.y <= 8)
end function

sub tChessGame.GenerateGeometricalMoves()
  GGM(moves(), movesNumber)
end sub

sub tChessGame.GenerateOpponentGeometricalMoves()
  activeColor = cBlack * activeColor
  GGM(opponentMoves(), opponentMovesNumber)
  activeColor = cBlack * activeColor
end sub

sub tChessGame.GGM(m() as tMove, byref mn as integer)
  dim i as integer
  dim as tVector a, b
  mn = 0
  for x as integer = 1 to 8
    for y as integer = 1 to 8
      if sgn(board(x, y)) = activeColor then
        a = type(x, y)
        select case as const abs(board(x, y))
          case cPawn
            for i = iif(activeColor = cWhite, 1, 4) to iif(activeColor = cWhite, 2, 5)
              b = a + pMove(i)
              if IsInBoard(b) andalso cbool(sgn(board(b.x, b.y)) = cBlack * activeColor) then
                mn += 1
                m(mn) = type(a, b, false)
              end if
            next i
            i = iif(activeColor = cWhite, 3, 6)
            b = a + pMove(i)
            if IsInBoard(b) andalso cbool(board(b.x, b.y) = cNothing) then
              mn += 1
              m(mn) = type(a, b, false)
              if a.y = iif(activeColor = cWhite, 2, 7) then
                b = b + pMove(i)
                if board(b.x, b.y) = cNothing then
                  mn += 1
                  m(mn) = type(a, b, false)
                end if
              end if
            end if
          case cKnight
            for i = 1 to 8
              b = a + nMove(i)
              if IsInBoard(b) andalso cbool(sgn(board(b.x, b.y)) <> activeColor) then
                mn += 1
                m(mn) = type(a, b, false)
              end if
            next i
          case cBishop
            for i = 1 to 4
              b = a + bMove(i)
              do while IsInBoard(b) andalso cbool(sgn(board(b.x, b.y)) <> activeColor)
                mn += 1
                m(mn) = type(a, b, false)
                if sgn(board(b.x, b.y)) = cBlack * activeColor then exit do
                b = b + bMove(i)
              loop
            next i
          case cRook
            for i = 1 to 4
              b = a + rMove(i)
              do while IsInBoard(b) andalso cbool(sgn(board(b.x, b.y)) <> activeColor)
                mn += 1
                m(mn) = type(a, b, false)
                if sgn(board(b.x, b.y)) = cBlack * activeColor then exit do
                b = b + rMove(i)
              loop
            next i
          case cQueen
            for i = 1 to 8
              b = a + qMove(i)
              do while IsInBoard(b) andalso cbool(sgn(board(b.x, b.y)) <> activeColor)
                mn += 1
                m(mn) = type(a, b, false)
                if sgn(board(b.x, b.y)) = cBlack * activeColor then exit do
                b = b + qMove(i)
              loop
            next i
          case cKing
            for i = 1 to 8
              b = a + kMove(i)
              if IsInBoard(b) andalso cbool(sgn(board(b.x, b.y)) <> activeColor) then
                mn += 1
                m(mn) = type(a, b, false)
              end if
            next i
        end select
      end if
    next y
  next x
end sub

function tChessGame.IsThreatened(v as tVector, doComputeMoves as boolean = false) as boolean
  dim result as boolean = false
  dim index as integer = 1 
  if doComputeMoves then GenerateOpponentGeometricalMoves()
  do while cbool(index <= opponentMovesNumber) andalso not result
    if opponentMoves(index).b = v then result = true
    index += 1
  loop
  function = result
end function

function tChessGame.Check() as boolean
  dim as integer x = 1, y = 1
  do while (board(x, y) <> cKing * activeColor) andalso (x * y <> 64)
    if x < 8 then
      x += 1
    else
      y += 1
      x = 1
    end if
  loop
  if board(x, y) <> cKing * activeColor then return false
  function = IsThreatened(type(x, y), true)
end function

sub tChessGame.GenerateSpecialMoves()
  dim as tVector a, b
  if enPassantTargetSquare.x <> 0 then
    b = enPassantTargetSquare
    a = type(b.x - 1, b.y - activeColor)
    if IsInBoard(a) andalso cbool(board(a.x, a.y) = activeColor * cPawn) then
      movesNumber += 1
      moves(movesNumber) = type(a, b, false)
    end if
    a = type(b.x + 1, b.y - activeColor)
    if IsInBoard(a) andalso cbool(board(a.x, a.y) = activeColor * cPawn) then
      movesNumber += 1
      moves(movesNumber) = type(a, b, false)
    end if
  end if
  if activeColor = cWhite then
    if board(5, 1) = activeColor * cKing then
      if board(8, 1) = activeColor * cRook then
        if (board(6, 1) = cNothing) andalso (board(7, 1) = cNothing) then
          if not IsThreatened(type(5, 1), true)_
          andalso not IsThreatened(type(6, 1), false)_
          andalso not IsThreatened(type(7, 1), false)_
          andalso not IsThreatened(type(8, 1), false) then
            if castlingAvailable(g1) then
              a = type(5, 1)
              b = type(7, 1)
              movesNumber += 1
              moves(movesNumber) = type(a, b, false)
            end if
          end if
        end if
      end if
      if board(1, 1) = activeColor * cRook then
        if (board(2, 1) = cNothing) andalso (board(3, 1) = cNothing) andalso (board(4, 1) = cNothing) then
          if not IsThreatened(type(1, 1), true)_
          andalso not IsThreatened(type(2, 1), false)_
          andalso not IsThreatened(type(3, 1), false)_
          andalso not IsThreatened(type(4, 1), false)_
          andalso not IsThreatened(type(5, 1), false) then
            if castlingAvailable(c1) then
              a = type(5, 1)
              b = type(3, 1)
              movesNumber += 1
              moves(movesNumber) = type(a, b, false)
            end if
          end if
        end if
      end if
    end if
  else
    if board(5, 8) = activeColor * cKing then
      if board(8, 8) = activeColor * cRook then
        if (board(6, 8) = cNothing) andalso (board(7, 8) = cNothing) then
          if not IsThreatened(type(5, 8), true)_
          andalso not IsThreatened(type(6, 8), false)_
          andalso not IsThreatened(type(7, 8), false)_
          andalso not IsThreatened(type(8, 8), false) then
            if castlingAvailable(g8) then
              a = type(5, 8)
              b = type(7, 8)
              movesNumber += 1
              moves(movesNumber) = type(a, b, false)
            end if
          end if
        end if
      end if
      if board(1, 8) = activeColor * cRook then
        if (board(2, 8) = cNothing) andalso (board(3, 8) = cNothing) andalso (board(4, 8) = cNothing) then
          if not IsThreatened(type(1, 8), true)_
          andalso not IsThreatened(type(2, 8), false)_
          andalso not IsThreatened(type(3, 8), false)_
          andalso not IsThreatened(type(4, 8), false)_
          andalso not IsThreatened(type(5, 8), false) then
            if castlingAvailable(c8) then
              a = type(5, 8)
              b = type(3, 8)
              movesNumber += 1
              moves(movesNumber) = type(a, b, false)
            end if
          end if
        end if
      end if
    end if
  end if
end sub

sub tChessGame.DoMove(m as tMove, promotion as integer = cQueen)
  dim as tVector a, b
  if (abs(board(m.a.x, m.a.y)) = cPawn) andalso (board(m.b.x, m.b.y) = cNothing) andalso (m.b.x <> m.a.x) then
    board(m.b.x, m.b.y - activeColor) = cNothing
  end if
  if (abs(board(m.a.x, m.a.y)) = cKing) andalso (abs(m.a.x - m.b.x) = 2) then
    select case m.b.x * m.b.y
      case 7 * 1
        a = type(8, 1)
        b = type(6, 1)
      case 3 * 1
        a = type(1, 1)
        b = type(4, 1)
      case 7 * 8
        a = type(8, 8)
        b = type(6, 8)
      case 3 * 8
        a = type(1, 8)
        b = type(4, 8)
    end select
    board(b.x, b.y) = board(a.x, a.y)
    board(a.x, a.y) = cNothing
  end if
  if (abs(board(m.a.x, m.a.y)) = cPawn) andalso (m.b.y = 1) or (m.b.y = 8) then
    board(m.a.x, m.a.y) *= promotion'cQueen
  end if
  if abs(board(m.a.x, m.a.y)) = cKing then
    if activeColor = cWhite then
      castlingAvailable(g1) = false
      castlingAvailable(c1) = false
    else
      castlingAvailable(g8) = false
      castlingAvailable(c8) = false
    end if
  end if
  if abs(board(m.a.x, m.a.y)) = cRook then
    if m.a.x = 8 then
      if m.a.y = 1 then castlingAvailable(g1) = false
      if m.a.y = 8 then castlingAvailable(g8) = false
    end if
    if m.a.x = 1 then
      if m.a.y = 1 then castlingAvailable(c1) = false
      if m.a.y = 8 then castlingAvailable(c8) = false
    end if
  end if
  if (abs(board(m.a.x, m.a.y)) = cPawn) andalso (abs(m.a.y - m.b.y) = 2) then
    with enPassantTargetSquare
      .x = m.a.x
      .y = m.a.y + activeColor
    end with
  else
    enPassantTargetSquare = type(0, 0)
  end if
  if (abs(board(m.a.x, m.a.y)) = cPawn) or (board(m.b.x, m.b.y) <> cNothing) then
    halfMoveClock = 0
  else
    halfMoveClock += 1
  end if
  if activeColor = cBlack then fullMoveNumber += 1
  board(m.b.x, m.b.y) = board(m.a.x, m.a.y)
  board(m.a.x, m.a.y) = cNothing
  activeColor *= cBlack
end sub

sub tChessGame.DoMove(m as string, promotion as integer = cQueen)
  dim m1 as tMove
  m1.a.x = m[0] - 96
  m1.a.y = m[1] - 48
  m1.b.x = m[2] - 96
  m1.b.y = m[3] - 48
  if len(m) = 5 then
    select case chr(m[4])
      case "n" : promotion = cKnight
      case "b" : promotion = cBishop
      case "r" : promotion = cRook
      case "q" : promotion = cQueen
    end select
  end if
  DoMove(m1, promotion)
end sub

sub tChessGame.CheckMovesLegality()
  dim as integer piecePlacement1(1 to 8, 1 to 8)
  dim as integer activeColor1
  dim as boolean castlingAvailable1(g1 to c8)
  dim as tVector enPassantTargetSquare1
  dim as integer halfmoveClock1
  dim as integer fullmoveNumber1
  dim as integer x, y, i
  dim as tCastlingTargetSquare c
  for x = 1 to 8
    for y = 1 to 8
      piecePlacement1(x, y) = board(x, y)
    next y
  next x
  activeColor1 = activeColor
  for c = g1 to c8
    castlingAvailable1(c) = castlingAvailable(c)
  next c
  with enPassantTargetSquare1
    .x = enPassantTargetSquare.x
    .y = enPassantTargetSquare.y
  end with
  halfmoveClock1 = halfmoveClock
  fullmoveNumber1 = fullmoveNumber
  for i = 1 to movesNumber
    DoMove(moves(i))
    activeColor *= cBlack
    moves(i).isLegal = not Check
    for x = 1 to 8
      for y = 1 to 8
        board(x, y) = piecePlacement1(x, y)
      next y
    next x
    activeColor = activeColor1
    for c = g1 to c8
      castlingAvailable(c) = castlingAvailable1(c)
    next c
    with enPassantTargetSquare
      .x = enPassantTargetSquare1.x
      .y = enPassantTargetSquare1.y
    end with
    halfmoveClock = halfmoveClock1
    fullmoveNumber = fullmoveNumber1
  next i
end sub

function tChessGame.IsLegal(m as string) as boolean
  dim result as boolean = false
  dim m1 as tMove
  dim i as integer = 1
  m1.a.x = m[0] - 96
  m1.a.y = m[1] - 48
  m1.b.x = m[2] - 96
  m1.b.y = m[3] - 48
  do while cbool(i <= movesNumber) andalso not result
    if (m1.a = moves(i).a) andalso (m1.b = moves(i).b) then result = moves(i).IsLegal
    i += 1
  loop
  function = result
end function

function tChessGame.IsLegalStartingSquare(sq as string) as boolean
  dim result as boolean = false
  dim v as tVector
  dim i as integer = 1
  v.x = sq[0] - 96
  v.y = sq[1] - 48
  do while cbool(i <= movesNumber) andalso not result
    if v = moves(i).a then result = moves(i).IsLegal
    i += 1
  loop
  function = result
end function

function tChessGame.LegalDestinations(sq as string) as string
  dim result as string = ""
  dim v as tVector
  dim i as integer = 1
  v.x = sq[0] - 96
  v.y = sq[1] - 48
  do while cbool(i <= movesNumber)
    if cbool(v = moves(i).a) and moves(i).IsLegal then
      result &= VectToStr(moves(i).b)
    end if
    i += 1
  loop
  function = result
end function

function tChessGame.GameState() as tGameState
  dim result as tGameState = gsInProgress
  dim as integer index = 1, legalMovesNumber = 0
  if halfMoveClock = 50 then result = gsDraw
  do while (index <= movesNumber) andalso (legalMovesNumber = 0)
    if moves(index).isLegal then legalMovesNumber += 1
    index += 1
  loop
  if legalMovesNumber = 0 then
    if Check then result = gsCheckmate else result = gsStalemate
  end if
  function = result
end function

function tChessGame.RandomMove() as integer
  dim result as integer = 0
  dim legalMoves() as integer
  dim legalMovesNumber as integer = 0
  for i as integer = 1 to movesNumber
    if moves(i).isLegal then
      legalMovesNumber += 1
      redim preserve legalMoves(1 to legalMovesNumber)
      legalMoves(legalMovesNumber) = i
    end if
  next i
  if legalMovesnumber > 0 then result = legalMoves(int(legalMovesNumber * rnd) + 1)
  function = result
end function

function tChessGame.FEN() as string
  dim s(1 to 6) as string = {"", "", "", "", "", ""}
  
  dim symbol as string = "kqrbnp PNBRQK"
  dim as integer x = 1, y = 8, z = 0
  
  do while y >= 1
    if board(x, y) = cNothing then
      do while (x <= 8) andalso (board(x, y) = cNothing)
        z += 1
        x += 1
      loop
      s(1) = s(1) & chr(z + 48)
      z = 0
    else
      s(1) = s(1) & chr(symbol[board(x, y) + cKing])
      x += 1
    end if
    if x = 9 then
      if y > 1 then s(1) = s(1) & "/"
      x = 1
      y -= 1
    end if
  loop
  
  if activeColor = cWhite then s(2) = "w" else s(2) = "b"
  
  z = 0
  for c as tCastlingTargetSquare = g1 to c8
    if castlingAvailable(c) then z += 1
  next c
  if z = 0 then
    s(3) = "-"
  else
    if castlingAvailable(g1) then s(3) = s(3) & "K"
    if castlingAvailable(c1) then s(3) = s(3) & "Q"
    if castlingAvailable(g8) then s(3) = s(3) & "k"
    if castlingAvailable(c8) then s(3) = s(3) & "q"
  end if
  
  if enPassantTargetSquare.x = 0 then
    s(4) = "-"
  else
    s(4) = VectToStr(enPassantTargetSquare)
  end if
  
  s(5) = str(halfmoveClock)
  s(6) = str(fullmoveNumber)
  
  function = s(1) & " " & s(2) & " " & s(3) & " " & s(4) & " " & s(5) & " " & s(6)
end function

function tChessGame.EPD() as string
  dim s(1 to 4) as string = {"", "", "", ""}
  
  dim symbol as string = "kqrbnp PNBRQK"
  dim as integer x = 1, y = 8, z = 0
  
  do while y >= 1
    if board(x, y) = cNothing then
      do while (x <= 8) andalso (board(x, y) = cNothing)
        z += 1
        x += 1
      loop
      s(1) = s(1) & chr(z + 48)
      z = 0
    else
      s(1) = s(1) & chr(symbol[board(x, y) + cKing])
      x += 1
    end if
    if x = 9 then
      if y > 1 then s(1) = s(1) & "/"
      x = 1
      y -= 1
    end if
  loop
  
  if activeColor = cWhite then s(2) = "w" else s(2) = "b"
  
  z = 0
  for c as tCastlingTargetSquare = g1 to c8
    if castlingAvailable(c) then z += 1
  next c
  if z = 0 then
    s(3) = "-"
  else
    if castlingAvailable(g1) then s(3) = s(3) & "K"
    if castlingAvailable(c1) then s(3) = s(3) & "Q"
    if castlingAvailable(g8) then s(3) = s(3) & "k"
    if castlingAvailable(c8) then s(3) = s(3) & "q"
  end if
  
  if enPassantTargetSquare.x = 0 then
    s(4) = "-"
  else
    s(4) = VectToStr(enPassantTargetSquare)
  end if
  
  function = s(1) & " " & s(2) & " " & s(3) & " " & s(4)
end function

end namespace
