
#define RandomInt(a, b) int((b - a + 1) * rnd + a)

namespace chess

type TVector
  as integer x
  as integer y
end type

type TMove
  as TVector a
  as TVector b
end type

operator + (v1 as TVector, v2 as TVector) as TVector
  return type(v1.x + v2.x, v1.y + v2.y)
end operator

operator = (v1 as TVector, v2 as TVector) as integer
  return (v1.x = v2.x) andalso (v1.y = v2.y)
end operator

function VecToStr(byval v as TVector) as string
  return chr(v.x + asc("a") - 1, v.y + asc("1") - 1)
end function

function StrToVec(byval s as string) as TVector
  return type(s[0] - asc("a") + 1, s[1] - asc("1") + 1)
end function

function MovToStr(byval m as TMove) as string
  return VecToStr(m.a) & VecToStr(m.b)
end function

function StrToMov(byval s as string) as TMove
  return type(StrToVec(left(s, 2)), StrToVec(mid(s, 3, 2)))
end function

const INITIAL_PLACEMENT = "rnbqkbnr/pppppppp/8/8/8/8/PPPPPPPP/RNBQKBNR w KQkq - 0 1"

const as integer _
  NOTHING = 0, _
  PAWN = 1, _
  KNIGHT = 2, _
  BISHOP = 3, _
  ROOK = 4, _
  QUEEN = 5, _
  KING = 6, _
  WHITE = 1, _
  BLACK = -1

const MAX_MOVESNUMBER = 100

type TGame
  as integer board(1 to 8, 1 to 8)
  as integer activeColor
  as string castlingAvailable
  as string enPassantTargetSquare
  as integer halfmoveClock
  as integer fullmoveNumber
  
  as TMove moves(1 to MAX_MOVESNUMBER)
  as integer movesNumber
  as TMove opponentMoves(1 to MAX_MOVESNUMBER)
  as integer opponentMovesNumber
  
  declare sub Initialize(byval aFen as string = INITIAL_PLACEMENT)
  declare function Fen() as string
  declare function IsInBoard(byval v as TVector) as integer
  declare sub GenerateMoves overload(byval aColor as integer, aMove() as TMove, byref aNumber as integer)
  declare sub GenerateMoves overload()
  
  declare function IsThreatened(byval v as TVector, byval computeMoves as integer = 0) as integer
  declare function Check() as integer
  declare sub GenerateSpecialMoves()
  
  declare sub RemoveCastlingPermission(byref aPermissionSet as string, byval aPermission as string)
  
  declare sub DoMove(byval m as TMove, byval promotion as integer = QUEEN)
  declare sub DoMove(byval m as string)
  declare sub SortLegalMoves()
  
  declare function FindMove(byval m as string) as integer
  declare function RandomMove() as string
  declare sub GenerateLegalMoves()
  
  declare function Text() as string
  declare function IsPromotion(byval m as TMove) as integer
end type

sub TGame.Initialize(byval aFen as string = INITIAL_PLACEMENT)
  dim as string s(1 to 6), symbols = "kqrbnp PNBRQK"
  dim as integer x, y, z
  ToLog(aFen)

  z = InStr(aFen, " ") : s(1) = left(aFen, z - 1) : aFen = right(aFen, len(aFen) - z)
  z = InStr(aFen, " ") : s(2) = left(aFen, z - 1) : aFen = right(aFen, len(aFen) - z)
  z = InStr(aFen, " ") : s(3) = left(aFen, z - 1) : aFen = right(aFen, len(aFen) - z)
  z = InStr(aFen, " ") : s(4) = left(aFen, z - 1) : aFen = right(aFen, len(aFen) - z)
  z = InStr(aFen, " ") : s(5) = left(aFen, z - 1) : aFen = right(aFen, len(aFen) - z)
  s(6) = aFen
  for i as integer = 1 to 6
    ToLog(s(i))
  next i

  x = 1
  y = 8
  
  for z = 1 to len(s(1))
    if s(1)[z - 1] = asc("/") then
      x = 1
      y -= 1
      continue for
    end if
    
    if (s(1)[z - 1] >= asc("1")) andalso (s(1)[z - 1] <= asc("8")) then
      do while s(1)[z - 1] > asc("0")
        board(x, y) = NOTHING
        s(1)[z - 1] -= 1
        x += 1
      loop
      continue for
    end if
    
    if InStr(symbols, chr(s(1)[z - 1])) > 0 then
      board(x, y) = InStr(symbols, chr(s(1)[z - 1])) - 7
      x += 1
      continue for
    end if
  next z
  
  activeColor = iif(s(2) = "w", WHITE, BLACK)
  castlingAvailable = s(3)
  enPassantTargetSquare = s(4)
  halfmoveClock = valint(s(5))
  fullmoveNumber = valint(s(6))
  
  ToLog(Fen())
end sub

function TGame.Fen() as string
  dim as string _
    result, _
    s(1 to 6) = {"", "", "", "", "", ""}, _
    symbol = "kqrbnp PNBRQK"
  dim as integer x = 1, y = 8, z = 0
  
  do while y >= 1
    if board(x, y) = NOTHING then
      do while (x <= 8) andalso (board(x, y) = NOTHING)
        z += 1
        x += 1
      loop
      s(1) = s(1) & chr(z + 48)
      z = 0
    else
      s(1) = s(1) & chr(symbol[board(x, y) + KING])
      x += 1
    end if
    if x = 9 then
      if y > 1 then s(1) = s(1) & "/"
      x = 1
      y -= 1
    end if
  loop
  
  s(2) = iif(activeColor = WHITE, "w", "b")
  s(3) = castlingAvailable
  s(4) = enPassantTargetSquare
  s(5) = str(halfmoveClock)
  s(6) = str(fullmoveNumber)
  
  result = s(1)
  for i as integer = 2 to 6
    result &= " " & s(i)
  next i
  return result
end function

function TGame.IsInBoard(byval v as TVector) as integer
  return (v.x >= 1) andalso (v.x <= 8) andalso (v.y >= 1) andalso (v.y <= 8)
end function

sub TGame.GenerateMoves(byval aColor as integer, aMove() as TMove, byref aNumber as integer)
  dim as TVector _
    PVEC(1 to 6), _
    NVEC(1 to 8), _
    BVEC(1 to 4), _
    RVEC(1 to 4), _
    QVEC(1 to 8), _
    KVEC(1 to 8)
  
  NVEC(1) = type(-1, +2)
  NVEC(2) = type(+1, +2)
  NVEC(3) = type(-2, +1)
  NVEC(4) = type(+2, +1)
  NVEC(5) = type(-2, -1)
  NVEC(6) = type(+2, -1)
  NVEC(7) = type(-1, -2)
  NVEC(8) = type(+1, -2)

  BVEC(1) = type(-1, +1)
  BVEC(2) = type(+1, +1)
  BVEC(3) = type(-1, -1)
  BVEC(4) = type(+1, -1)

  RVEC(1) = type(0, +1)
  RVEC(2) = type(0, -1)
  RVEC(3) = type(-1,  0)
  RVEC(4) = type(+1,  0)

  KVEC(1) = BVEC(1)
  KVEC(2) = BVEC(2)
  KVEC(3) = BVEC(3)
  KVEC(4) = BVEC(4)
  KVEC(5) = RVEC(1)
  KVEC(6) = RVEC(2)
  KVEC(7) = RVEC(3)
  KVEC(8) = RVEC(4)

  QVEC(1) = KVEC(1)
  QVEC(2) = KVEC(2)
  QVEC(3) = KVEC(3)
  QVEC(4) = KVEC(4)
  QVEC(5) = KVEC(5)
  QVEC(6) = KVEC(6)
  QVEC(7) = KVEC(7)
  QVEC(8) = KVEC(8)

  PVEC(1) = BVEC(1)
  PVEC(2) = BVEC(2)
  PVEC(4) = BVEC(3)
  PVEC(5) = BVEC(4)
  PVEC(3) = RVEC(1)
  PVEC(6) = RVEC(2)
  
  dim i as integer
  dim as TVector a, b
  
  for x as integer = 1 to 8
    for y as integer = 1 to 8
      if sgn(board(x, y)) = aColor then
        a = type(x, y)
        select case as const abs(board(x, y))
          case PAWN
            for i = iif(aColor = WHITE, 1, 4) to iif(aColor = WHITE, 2, 5)
              b = a + PVEC(i)
              if IsInBoard(b) andalso (sgn(board(b.x, b.y)) = BLACK * aColor) then
                aNumber += 1
                aMove(aNumber) = type(a, b)
              end if
            next i
            i = iif(aColor = WHITE, 3, 6)
            b = a + PVEC(i)
            if IsInBoard(b) andalso (board(b.x, b.y) = NOTHING) then
              aNumber += 1
              aMove(aNumber) = type(a, b)
              if a.y = iif(aColor = WHITE, 2, 7) then
                b = b + PVEC(i)
                if board(b.x, b.y) = NOTHING then
                  aNumber += 1
                  aMove(aNumber) = type(a, b)
                end if
              end if
            end if
          case KNIGHT
            for i = 1 to 8
              b = a + NVEC(i)
              if IsInBoard(b) andalso (sgn(board(b.x, b.y)) <> aColor) then
                aNumber += 1
                aMove(aNumber) = type(a, b)
              end if
            next i
          case BISHOP
            for i = 1 to 4
              b = a + BVEC(i)
              do while IsInBoard(b) andalso (sgn(board(b.x, b.y)) <> aColor)
                aNumber += 1
                aMove(aNumber) = type(a, b)
                if sgn(board(b.x, b.y)) = BLACK * aColor then exit do
                b = b + BVEC(i)
              loop
            next i
          case ROOK
            for i = 1 to 4
              b = a + RVEC(i)
              do while IsInBoard(b) andalso (sgn(board(b.x, b.y)) <> aColor)
                aNumber += 1
                aMove(aNumber) = type(a, b)
                if sgn(board(b.x, b.y)) = BLACK * aColor then exit do
                b = b + RVEC(i)
              loop
            next i
          case QUEEN
            for i = 1 to 8
              b = a + QVEC(i)
              do while IsInBoard(b) andalso (sgn(board(b.x, b.y)) <> aColor)
                aNumber += 1
                aMove(aNumber) = type(a, b)
                if sgn(board(b.x, b.y)) = BLACK * aColor then exit do
                b = b + QVEC(i)
              loop
            next i
          case KING
            for i = 1 to 8
              b = a + KVEC(i)
              if IsInBoard(b) andalso (sgn(board(b.x, b.y)) <> aColor) then
                aNumber += 1
                aMove(aNumber) = type(a, b)
              end if
            next i
        end select
      end if
    next y
  next x
end sub

sub TGame.GenerateMoves()
  GenerateMoves(activeColor, moves(), movesNumber)
end sub

function TGame.IsThreatened(byval v as TVector, byval computeMoves as integer = 0) as integer
  dim result as integer = 0
  dim index as integer = 1
  
  if computeMoves then
    opponentMovesNumber = 0
    GenerateMoves(BLACK * activeColor, opponentMoves(), opponentMovesNumber)
  end if
  
  do while (index <= opponentMovesNumber) andalso not result
    if opponentMoves(index).b = v then result = -1
    index += 1
  loop
  
  return result
end function

function TGame.Check() as integer
  dim as integer x = 1, y = 1
  do while (board(x, y) <> KING * activeColor) andalso (x * y <> 64)
    if x < 8 then
      x += 1
    else
      y += 1
      x = 1
    end if
  loop
  if board(x, y) <> KING * activeColor then return 0
  return IsThreatened(type(x, y), -1)
end function

sub TGame.GenerateSpecialMoves()
  dim as TVector a, b
  
  if enPassantTargetSquare <> "-" then
    ToLog(enPassantTargetSquare)
    b = StrToVec(enPassantTargetSquare)
    
    a = type(b.x - 1, b.y - activeColor)
    if IsInBoard(a) andalso (board(a.x, a.y) = activeColor * PAWN) then
      movesNumber += 1
      moves(movesNumber) = type(a, b)
      ToLog(MovToStr(moves(movesNumber)))
    end if
    
    a = type(b.x + 1, b.y - activeColor)
    if IsInBoard(a) andalso (board(a.x, a.y) = activeColor * PAWN) then
      movesNumber += 1
      moves(movesNumber) = type(a, b)
      ToLog(MovToStr(moves(movesNumber)))
    end if
  end if
  
  if activeColor = WHITE then
    if board(5, 1) = activeColor * KING then
      if board(8, 1) = activeColor * ROOK then
        if (board(6, 1) = NOTHING) andalso (board(7, 1) = NOTHING) then
          if not IsThreatened(type(5, 1), -1)_
          andalso not IsThreatened(type(6, 1), 0)_
          andalso not IsThreatened(type(7, 1), 0)_
          andalso not IsThreatened(type(8, 1), 0) then
            if InStr(castlingAvailable, "K") then
              a = type(5, 1)
              b = type(7, 1)
              movesNumber += 1
              moves(movesNumber) = type(a, b)
            end if
          end if
        end if
      end if
      if board(1, 1) = activeColor * ROOK then
        if (board(2, 1) = NOTHING) andalso (board(3, 1) = NOTHING) andalso (board(4, 1) = NOTHING) then
          if not IsThreatened(type(1, 1), -1)_
          andalso not IsThreatened(type(2, 1), 0)_
          andalso not IsThreatened(type(3, 1), 0)_
          andalso not IsThreatened(type(4, 1), 0)_
          andalso not IsThreatened(type(5, 1), 0) then
            if InStr(castlingAvailable, "Q") then
              a = type(5, 1)
              b = type(3, 1)
              movesNumber += 1
              moves(movesNumber) = type(a, b)
            end if
          end if
        end if
      end if
    end if
  else
    if board(5, 8) = activeColor * KING then
      if board(8, 8) = activeColor * ROOK then
        if (board(6, 8) = NOTHING) andalso (board(7, 8) = NOTHING) then
          if not IsThreatened(type(5, 8), -1)_
          andalso not IsThreatened(type(6, 8), 0)_
          andalso not IsThreatened(type(7, 8), 0)_
          andalso not IsThreatened(type(8, 8), 0) then
            if InStr(castlingAvailable, "k") then
              a = type(5, 8)
              b = type(7, 8)
              movesNumber += 1
              moves(movesNumber) = type(a, b)
            end if
          end if
        end if
      end if
      if board(1, 8) = activeColor * ROOK then
        if (board(2, 8) = NOTHING) andalso (board(3, 8) = NOTHING) andalso (board(4, 8) = NOTHING) then
          if not IsThreatened(type(1, 8), -1)_
          andalso not IsThreatened(type(2, 8), 0)_
          andalso not IsThreatened(type(3, 8), 0)_
          andalso not IsThreatened(type(4, 8), 0)_
          andalso not IsThreatened(type(5, 8), 0) then
            if InStr(castlingAvailable, "q") then
              a = type(5, 8)
              b = type(3, 8)
              movesNumber += 1
              moves(movesNumber) = type(a, b)
            end if
          end if
        end if
      end if
    end if
  end if
end sub

sub TGame.RemoveCastlingPermission(byref aPermissionSet as string, byval aPermission as string)
  dim i as integer = InStr(aPermissionSet, aPermission)
  if i <> 0 then
    if len(aPermissionSet) = 1 then
      aPermissionSet = "-"
    else
      aPermissionSet = left(aPermissionSet, i - 1) & right(aPermissionSet, len(aPermissionSet) - i) 
    end if
  end if
end sub

sub TGame.DoMove(byval m as TMove, byval promotion as integer = QUEEN)
  dim as tVector a, b
  if (abs(board(m.a.x, m.a.y)) = PAWN) andalso (board(m.b.x, m.b.y) = NOTHING) andalso (m.b.x <> m.a.x) then
    board(m.b.x, m.b.y - activeColor) = NOTHING
  end if
  if (abs(board(m.a.x, m.a.y)) = KING) andalso (abs(m.a.x - m.b.x) = 2) then
    select case m.b.x * m.b.y
      case 7 * 1
        a = type(8, 1)
        b = type(6, 1)
      case 3 * 1
        a = type(1, 1)
        b = type(4, 1)
      case 7 * 8
        a = type(8, 8)
        b = type(6, 8)
      case 3 * 8
        a = type(1, 8)
        b = type(4, 8)
    end select
    board(b.x, b.y) = board(a.x, a.y)
    board(a.x, a.y) = NOTHING
  end if
  if (abs(board(m.a.x, m.a.y)) = PAWN) andalso (m.b.y = 1) or (m.b.y = 8) then
    board(m.a.x, m.a.y) *= promotion
  end if
  if abs(board(m.a.x, m.a.y)) = KING then
    if activeColor = WHITE then
      RemoveCastlingPermission(castlingAvailable, "K")
      RemoveCastlingPermission(castlingAvailable, "Q")
    else
      RemoveCastlingPermission(castlingAvailable, "k")
      RemoveCastlingPermission(castlingAvailable, "q")
    end if
  end if
  if abs(board(m.a.x, m.a.y)) = ROOK then
    if m.a.x = 8 then
      if m.a.y = 1 then RemoveCastlingPermission(castlingAvailable, "K")
      if m.a.y = 8 then RemoveCastlingPermission(castlingAvailable, "k")
    end if
    if m.a.x = 1 then
      if m.a.y = 1 then RemoveCastlingPermission(castlingAvailable, "Q")
      if m.a.y = 8 then RemoveCastlingPermission(castlingAvailable, "q")
    end if
  end if
  if (abs(board(m.a.x, m.a.y)) = PAWN) andalso (abs(m.a.y - m.b.y) = 2) then
    enPassantTargetSquare = VecToStr(type(m.a.x, m.a.y + activeColor))
  else
    enPassantTargetSquare = "-"
  end if
  if (abs(board(m.a.x, m.a.y)) = PAWN) or (board(m.b.x, m.b.y) <> NOTHING) then
    halfMoveClock = 0
  else
    halfMoveClock += 1
  end if
  if activeColor = BLACK then fullMoveNumber += 1
  board(m.b.x, m.b.y) = board(m.a.x, m.a.y)
  board(m.a.x, m.a.y) = NOTHING
  activeColor *= BLACK
end sub

sub TGame.DoMove(byval m as string)
  dim promotion as integer
  
  if len(m) = 5 then
    select case chr(m[4])
      case "n"
        promotion = KNIGHT
      case "b"
        promotion = BISHOP
      case "r"
        promotion = ROOK
      case "q"
        promotion = QUEEN
    end select
    DoMove(StrToMov(m), promotion)
  else
    DoMove(StrToMov(m))
  end if
  
  ToLog("[" & m & "]")
end sub

sub TGame.SortLegalMoves()
  dim as integer piecePlacement1(1 to 8, 1 to 8)
  dim as integer activeColor1
  dim as string castlingAvailable1
  dim as string enPassantTargetSquare1
  dim as integer halfmoveClock1
  dim as integer fullmoveNumber1
  dim as integer x, y, i
  
  for x = 1 to 8
    for y = 1 to 8
      piecePlacement1(x, y) = board(x, y)
    next y
  next x
  activeColor1 = activeColor
  castlingAvailable1 = castlingAvailable
  enPassantTargetSquare1 = enPassantTargetSquare
  halfmoveClock1 = halfmoveClock
  fullmoveNumber1 = fullmoveNumber
  
  i = 1
  do while i <= movesNumber
    DoMove(moves(i))
    activeColor *= BLACK
    
    if Check() then
      ToLog("[" & str(i) & "][" & MovToStr(moves(i)) & "]")
      swap moves(i), moves(movesNumber)
      movesNumber -= 1
    else
      i += 1
    end if
    
    for x = 1 to 8
      for y = 1 to 8
        board(x, y) = piecePlacement1(x, y)
      next y
    next x
    activeColor = activeColor1
    castlingAvailable = castlingAvailable1
    enPassantTargetSquare = enPassantTargetSquare1
    halfmoveClock = halfmoveClock1
    fullmoveNumber = fullmoveNumber1
  loop
end sub

function TGame.FindMove(byval m as string) as integer
  dim result as integer = 0
  dim m1 as TMove
  dim i as integer = 1
  m1 = StrToMov(m)
  do while (i <= movesNumber) andalso not result
    if (m1.a = moves(i).a) andalso (m1.b = moves(i).b) then result = -1
    i += 1
  loop
  return result
end function

function TGame.RandomMove() as string
  dim i as integer
  if movesNumber = 0 then
    return "0000"
  else
    i = RandomInt(1, movesNumber)
    if IsPromotion(moves(i)) then
      return MovToStr(moves(i)) & mid("nbrq", RandomInt(1, 4), 1)
    else
      return MovToStr(moves(i))
    end if
  end if
end function

sub TGame.GenerateLegalMoves()
  movesNumber = 0
  GenerateMoves()
  GenerateSpecialMoves()
  SortLegalMoves()
end sub

function TGame.Text() as string
  const letters = "+ A B C D E F G H"
  dim symbol as string = "kqrbnp-PNBRQK"
  dim result as string = letters & chr(13, 10)
  for y as integer = 8 to 1 step -1
    result += chr(y + 48)
    for x as integer = 1 to 8
      result += chr(32, symbol[board(x, y) + KING])
    next x
    result += chr(13, 10)
  next y
  result += letters
  return result
end function

function TGame.IsPromotion(byval m as TMove) as integer
  return (abs(board(m.a.x, m.a.y)) = PAWN) and ((m.b.y = 1) or (m.b.y = 8))
end function

end namespace
